/*
 * barcode.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

#define copy_barcode(dst,src)                              \
do {                                                       \
    unit_t *from = src;                                    \
    unit_t *from_end = from + units_per_barcode;           \
    unit_t *to = dst;                                      \
    for (; from < from_end; from++,to++) { *to = *from; }  \
} while (0)

#define read_barcode(barcode,pos) \
    ((((unit_t*)barcode)[pos / BP_PER_UNIT] >> ((BP_PER_UNIT-1-pos%BP_PER_UNIT)*2)) & 3)

#define write_barcode(barcode,pos,binnuc)                                                                          \
do {                                                                                                               \
    unsigned unit_index = (pos) / BP_PER_UNIT;                                                                     \
    unsigned pos_in_unit = (pos) % BP_PER_UNIT;                                                                    \
    unsigned shift = (BP_PER_UNIT - 1 - pos_in_unit) * 2;                                                          \
    unsigned long long mask = ~(3ull << shift);                                                                    \
    ((unit_t*)(barcode))[unit_index] = (((unit_t*)(barcode))[unit_index] & mask) | (((unit_t)(binnuc)) << shift);  \
} while (0)

static int compare_barcodes(const void *b1,const void *b2)
{
    const void *b1_end = (unit_t*)b1 + units_per_barcode;

    while (b1 < b1_end)
    {
        if (*(unit_t*)b1 != *(unit_t*)b2) { return (*(unit_t*)b1 < *(unit_t*)b2) ? -1 : 1; }
        b1 = (unit_t*)b1 + 1;
        b2 = (unit_t*)b2 + 1;
    }

    return 0;
}

static int same_barcodes(const unit_t *b1,const unit_t *b2)
{
    const unit_t *b1_end = b1 + units_per_barcode;
    for (; b1 < b1_end; b1++,b2++) { if (*b1 != *b2) { return 0; } }
    return 1;
}

void print_complete_barcode(unit_t *b)
{
    int p, j = -1, shift;
    unsigned used = BP_PER_UNIT;
    unit_t u;
    for (p=0; p<template_length; p++)
    {
        if (use_barcode_position[p])
        {
            if (used >= BP_PER_UNIT) { j++; u = b[j]; used = 0; shift = 62; }
            putchar(binary_to_nucleotide[(u >> shift) & 3]);
            shift -= 2;
            used++;
        }
        else { putchar(template[p]); }
    }
}

#endif
