/*
 * cluster.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

unit_t* find_barcode(unit_t *b)
{
    unsigned bi = (*b) >> INDEX_SHIFT;
    return bsearch(b,clustered_barcode_index[bi],volume_sizes[bi],cluster_entry_size,&compare_barcodes);
}

void create_clusters()
{
    long long n = 0ll;
    void *p;
    int i,j;

    fprintf(stderr,"Creating clusters ..");

    clustered_barcode_index = (void**)calloc1(N_VOLUMES,sizeof(void*));
    cluster_entries = (void*)malloc1(cluster_entry_size*n_unique_barcodes);
    p = cluster_entries;

    for (i=0; i<N_VOLUMES; i++)
    {
        clustered_barcode_index[i] = p;
        for (j=0; j<volume_sizes[i]; j++)
        {
            cluster_entry_header_t *h = (cluster_entry_header_t*)(p + bytes_per_barcode);
            copy_barcode(p, index[i] + j*units_per_barcode);
            h->next = p;
            h->cluster_id = n;
            n++;
            p += cluster_entry_size;
        }
    }

    if (index)
    {
        for (i=0; i<N_VOLUMES; i++)
        {
            if (index[i]) { free1(index[i],volume_allocated_sizes[i]); }
        }
        free1(index,sizeof(unit_t*)*N_VOLUMES);
        index = 0;
    }
    if (volume_allocated_sizes) { free1(volume_allocated_sizes,sizeof(*volume_allocated_sizes)*N_VOLUMES); volume_allocated_sizes = 0; }

    /*fprintf(stderr," OK (now %lld bytes allocated)\n",allocated);*/
    fprintf(stderr," OK\n");
}

/*
 * Generates the next combination.
 * Returns 1 on success, 0 when there is no next combination.
 */
unsigned next_combination(unsigned *c,const unsigned n,const int k)
{
    int i;
    if (c[k-1] < n-1) { c[k-1]++; return 1; }
    for (i=k-2; i>=0; i--) { if (c[i] < n-k+i) { break; } }
    if (i < 0) { return 0; }
    c[i]++;
    for (; i<k-1; i++) { c[i+1] = c[i]+1; }
    return 1;
}

void cluster(unsigned d)
{
    int i;
    unit_t m[units_per_barcode];
    unsigned n_nuc_changes = 1 << (d*2);
    long long n_joined = 0ll;

    fprintf(stderr,"Clustering (with distance %u) ",d);
    for (i=0; i<n_unique_barcodes; i++)
    {
        unsigned p[d];
        unsigned nuc_change;
        unit_t *barcode = cluster_entries + cluster_entry_size*i;
        cluster_entry_header_t *header = (cluster_entry_header_t*)((void*)barcode + bytes_per_barcode);

        if (i%5000==0) { fprintf(stderr,"."); }

        /*printf("\n\n"); print_complete_barcode(cluster_entries + cluster_entry_size*i); printf("\n");*/

        { unsigned pi; for (pi=0; pi<d; pi++) { p[pi] = pi; } }
        do {
            for (nuc_change=0; nuc_change<n_nuc_changes; nuc_change++)
            {
                unsigned j;
                unsigned change = nuc_change;
                unit_t* f;
                cluster_entry_header_t *h;

                copy_barcode(m,cluster_entries + cluster_entry_size*i);
                for (j=0; j<d; j++) { write_barcode(m,p[j],change&3); change >>= 2; }
                if (compare_barcodes(barcode,m) >= 0) { continue; }
                /*print_complete_barcode(m); printf("\n");*/

                f = find_barcode(m);
                if (!f) { continue; }

                h = (cluster_entry_header_t*)((void*)f + bytes_per_barcode);
                if (h->cluster_id != header->cluster_id)
                {
                    cluster_entry_header_t *from_header,*to_header;
                    unit_t *from_barcode,*to_barcode;
                    if (h->cluster_id < header->cluster_id) { from_header = header; to_header = h; from_barcode = barcode; to_barcode = f; }
                    else { from_header = h; to_header = header; from_barcode = f; to_barcode = barcode; }

                    /*printf("Joining:\n");
                    printf("    @%20llu @%20llu ",(unsigned long long)to_barcode,(unsigned long long)to_header);
                    print_complete_barcode(to_barcode); printf("  next: %20llu, id: %lld\n",(unsigned long long)to_header->next,to_header->cluster_id);
                    printf("    @%20llu @%20llu ",(unsigned long long)from_barcode,(unsigned long long)from_header);
                    print_complete_barcode(from_barcode); printf("  next: %20llu, id: %lld\n",(unsigned long long)from_header->next,from_header->cluster_id);*/

                    if (from_header->next == from_barcode && to_header->next == to_barcode)
                    {
                        from_header->cluster_id = to_header->cluster_id;
                        from_header->next = to_barcode;
                        to_header->next = from_barcode;
                    }
                    else if (from_header->next == from_barcode)
                    {
                        from_header->cluster_id = to_header->cluster_id;
                        from_header->next = to_header->next;
                        to_header->next = from_barcode;
                    }
                    else if (to_header->next == to_barcode)
                    {
                        {
                            void *t = from_barcode;
                            do {
                                ((cluster_entry_header_t*)(t+bytes_per_barcode))->cluster_id = to_header->cluster_id;
                                t = ((cluster_entry_header_t*)(t+bytes_per_barcode))->next;
                            } while (t != from_barcode);
                        }
                        to_header->next = from_header->next;
                        from_header->next = to_barcode;
                    }
                    else
                    {
                        {
                            void *t = from_barcode;
                            do {
                                ((cluster_entry_header_t*)(t+bytes_per_barcode))->cluster_id = to_header->cluster_id;
                                t = ((cluster_entry_header_t*)(t+bytes_per_barcode))->next;
                            } while (t != from_barcode);
                        }
                        unit_t *t = from_header->next;
                        from_header->next = to_header->next;
                        to_header->next = t;
                    }
                    from_header->cluster_id = to_header->cluster_id;
                    n_joined++;
                }
            }
        } while (next_combination(p,barcode_complexity,d));

        /*return;*/

        /*copy_barcode(m,cluster_entries + cluster_entry_size*i);*/

        /*{
            unit_t c[units_per_barcode];
            unit_t *cc;
            int j;
            for (j=0; j<units_per_barcode; j++) { c[j] = 0ull; }
            for (j=0; j<barcode_complexity; j++)
            {
                unsigned bn = read_barcode(m,j);
                write_barcode(c,j,bn);
            }

            if (!same_barcodes(m,c))
            {
                fprintf(stderr,"%20llu %20llu != %20llu %20llu\n",m[0],m[1],c[0],c[1]);
            }

            cc = find_barcode(c);
            if (!cc || !same_barcodes(cc,m))
            {
                fprintf(stderr,"Can't find barcode %llu %llu\n",m[0],m[1]);
            }
        }*/
    }
    fprintf(stderr," OK (%lld joined)\n",n_joined);
}

void cluster_all()
{
    unsigned d;
    for (d=1; d<=dist; d++) { cluster(d); }
}

void count_clusters()
{
    long long n_clusters = 0ll;
    int i;
    for (i=0; i<n_unique_barcodes; i++)
    {
        unit_t *barcode = cluster_entries + cluster_entry_size*i;
        cluster_entry_header_t *header = (cluster_entry_header_t*)((void*)barcode + bytes_per_barcode);
        if (header->cluster_id == i) { n_clusters++; }
    }
    printf("%lld clusters\n",n_clusters);
}

#endif
