/*
 * defines.h
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifndef CLUSTERIZER_DEFINES_H
#define CLUSTERIZER_DEFINES_H

#define PROGRAM_NAME      "Nucleotide Sequence Clusterizer"
#define EXECUTABLE_NAME   "nucleotide-sequence-clusterizer"
#define VERSION           "0.0.1"
#define RELEASE_DATE      "2014-04-05"
#define COPYRIGHT_YEARS   "2014"
#define COPYRIGHT_AUTHORS "Kirill Kryukov"

/*
 * Each barcode is stored in binary in a series of units.
 * One unit is normally platform's largest native integer.
 */
#define unit_t unsigned long long

/*
 * Metrics derived from unit type.
 */
#define BYTES_PER_UNIT sizeof(unit_t)                /* 8 */
#define BP_PER_BYTE 4
#define BP_PER_UNIT (BYTES_PER_UNIT * BP_PER_BYTE)   /* 32 */

/*
 * How many units in one block.
 */
#define UNITS_PER_BLOCK (BYTES_PER_BLOCK / BYTES_PER_UNIT)

/*
 * How much a barcode unit should be right-shifted to obtain index.
 */
#define INDEX_SHIFT ((BP_PER_UNIT - INDEX_DEPTH) * 2)

/*
 * How many volumes are created by index.
 */
#define N_VOLUMES (1 << (INDEX_DEPTH * 2))

#endif
